# Copyright (C) 2021 Matthias Nadig

import numpy as np


def _convert_se_to_cw(bounds):
    """
    Convert from (start, end) to (center, width)
    """

    bounds_new = np.empty(bounds.shape)
    bounds_new[..., 1] = np.subtract(bounds[..., 1], bounds[..., 0], out=bounds_new[..., 1])
    bounds_new[..., 0] = np.add(bounds[..., 0], np.multiply(0.5, bounds_new[..., 1]), out=bounds_new[..., 0])

    return bounds_new


def _convert_cw_to_se(bounds):
    """
    Convert from (center, width) to (start, end)
    """

    bounds_new = np.empty(bounds.shape)

    bounds_center = np.copy(bounds[..., 0])
    bounds_size_half = np.multiply(0.5, bounds[..., 1])

    bounds_new[..., 0] = np.subtract(bounds_center, bounds_size_half)
    bounds_new[..., 1] = np.add(bounds_center, bounds_size_half)

    return bounds_new


def _convert_se_to_cw_inplace(bounds):
    """
    Convert from (start, end) to (center, width)
    CAVEAT: Inplace modification!
    """

    np.subtract(bounds[..., 1], bounds[..., 0], out=bounds[..., 1])
    np.add(bounds[..., 0], np.multiply(0.5, bounds[..., 1]), out=bounds[..., 0])

    return bounds


def _convert_cw_to_se_inplace(bounds):
    """
    Convert from (center, width) to (start, end)
    CAVEAT: Inplace modification!
    """

    bounds_center = np.copy(bounds[..., 0])
    bounds_size_half = np.multiply(0.5, bounds[..., 1], out=bounds[..., 1])

    np.subtract(bounds_center, bounds_size_half, out=bounds[..., 0])
    np.add(bounds_center, bounds_size_half, out=bounds[..., 1])

    return bounds



# Copyright (C) 2021 Matthias Nadig

import numpy as np

from ._conversion import _convert_cw_to_se
from ._conversion import _convert_cw_to_se_inplace
from ._conversion import _convert_se_to_cw
from ._conversion import _convert_se_to_cw_inplace


class NdBounds:
    """
    n-Dimensional Bounds

    Basic format is:
        [[start dim 1, end dim 1],
         [start dim 2, end dim 2],
         ...,
         [start dim n, end dim n]]
    """

    def __init__(self, bounds):
        self._bound_handler = bounds

    def get_n_dims(self):
        """
        Returns the number of dimensions of the bounds
        (NOT OF THE NUMPY ARRAY ITSELF)
        """
        return self._bound_handler.get_n_dims()

    def get_shape(self):
        """ Returns the shape of the Numpy array that contains the bounds """
        return self._bound_handler.get_shape()

    def get_bounds_se(self):
        """ Returns the bounds in a Numpy array with format (start, end) """
        return self._bound_handler.get_bounds_se()

    def get_bounds_cw(self):
        """ Returns the bounds in a Numpy array with format (center, width) """
        return self._bound_handler.get_bounds_cw()

    def convert_to_se(self):
        """
        Makes sure, that the bounds are kept in format (start, end)
        Note: This method is meant for runtime improvements only, and
        can be called, before the bounds are retrieved in SE format
        multiple times.
        """
        if isinstance(self._bound_handler, NdBoundsSE):
            pass
        elif isinstance(self._bound_handler, NdBoundsCW):
            bounds = self._bound_handler.get_bounds_cw()
            _convert_cw_to_se_inplace(bounds)
            self._bound_handler = NdBoundsSE(bounds)
        else:
            raise RuntimeError(type(self._bound_handler))

        return self

    def convert_to_cw(self):
        """
        Makes sure, that the bounds are kept in format (center, width)
        Also see convert_to_se().
        """
        if isinstance(self._bound_handler, NdBoundsCW):
            pass
        elif isinstance(self._bound_handler, NdBoundsSE):
            bounds = self._bound_handler.get_bounds_se()
            _convert_se_to_cw_inplace(bounds)
            self._bound_handler = NdBoundsCW(bounds)
        else:
            raise RuntimeError(type(self._bound_handler))

        return self

    def scale_dimensions(self, factor_each_dim):
        """ Scales the dimensions, the bounds refer to """
        self._bound_handler.scale_dimensions(factor_each_dim)
        return self

    def scale_width(self, factor_each_dim):
        """ Scales the width of the individual bounds """
        self._bound_handler.scale_width(factor_each_dim)
        return self

    def add_offset(self, offset_each_dim):
        """ Adds offset to the individual bounds """
        self._bound_handler.add_offset(offset_each_dim)
        return self

    def check_within(self, point):
        """ Returns True or False for each of the bounds, depending on if the point lies within them """

        # Check input
        point = np.asarray(point)
        if point.ndim != 1 or point.shape[-1] != self.get_n_dims():
            raise ValueError('point needs to be {}-dimensional, '.format(self.get_n_dims()) +
                             'got point with shape {}'.format(point.shape))

        # Repeat to fit shape of bounds
        for index_axis, n_repeat in enumerate(self.get_shape()[:-2]):
            point = np.expand_dims(point, axis=index_axis)
            point = np.repeat(point, n_repeat, axis=index_axis)

        # Check if point is within start and end
        bounds = self.get_bounds_se()
        is_after_start = bounds[..., 0] <= point
        is_before_end = bounds[..., 1] >= point
        is_within_bounds = np.logical_and(is_after_start, is_before_end).all(axis=-1)

        return is_within_bounds


class BoundFormatHandler:
    """
    A class that serves as interface to the actual array of bounds
    """

    def __init__(self, bounds):
        self._bounds = bounds.astype(np.float)

    def get_n_dims(self):
        return self._bounds.shape[-2]

    def get_shape(self):
        return self._bounds.shape

    def get_bounds_se(self):
        return self._get_bounds_se()

    def get_bounds_cw(self):
        return self._get_bounds_cw()

    def scale_dimensions(self, factor_each_dim):
        factor_each_dim = self._reshape_input_for_application_to_each_bounds(factor_each_dim)
        return self._scale_dimensions(factor_each_dim)

    def scale_width(self, factor_each_dim):
        factor_each_dim = self._reshape_input_for_application_to_each_bounds(factor_each_dim)
        return self._scale_width(factor_each_dim)

    def add_offset(self, offset_each_dim):
        offset_each_dim = self._reshape_input_for_application_to_each_bounds(offset_each_dim)
        return self._add_offset(offset_each_dim)

    def _get_bounds_se(self):
        # Should be overwritten by child
        raise RuntimeError()

    def _get_bounds_cw(self):
        # Should be overwritten by child
        raise RuntimeError()

    def _set_bounds_from_se(self, bounds_se):
        # Should be overwritten by child
        raise RuntimeError()

    def _set_bounds_from_cw(self, bounds_cw):
        # Should be overwritten by child
        raise RuntimeError()

    def _add_offset(self, offset_each_dim):
        # Should be overwritten by child
        raise RuntimeError()

    def _reshape_input_for_application_to_each_bounds(self, arr):
        arr = np.asarray(arr)
        if arr.ndim == 1 and arr.shape[-1] == self.get_n_dims():
            # Repeat the array for each bounds
            for n_elements in reversed(self.get_shape()[:-2]):
                arr = np.repeat(arr[np.newaxis], n_elements, axis=0)
        elif arr.shape == self.get_shape()[:-1]:
            # Shape already as required for calculations
            pass
        else:
            raise ValueError('Bad shape for input array. Expected {}, got {}.'.format(
                '{} or {}'.format(self.get_shape()[-2:-1], self.get_shape()[:-1]),
                arr.shape
            ))

        return arr

    def _scale_dimensions(self, factor_each_dim):
        # Apply scaling factors
        bounds_se = self.get_bounds_se()
        bounds_se = np.multiply(bounds_se, factor_each_dim[..., np.newaxis], out=bounds_se)

        # Convert result
        self._set_bounds_from_se(bounds_se)

        return self

    def _scale_width(self, factor_each_dim):
        # Apply scaling factors
        bounds_cw = self.get_bounds_cw()
        np.multiply(bounds_cw[..., 1], factor_each_dim, out=bounds_cw[..., 1])

        # Convert result
        self._set_bounds_from_cw(bounds_cw)

        return self


class NdBoundsSE(BoundFormatHandler):
    """ Handler for the format (start, end) """

    def __init__(self, bounds):
        super().__init__(bounds)

    def _get_bounds_se(self):
        return self._bounds

    def _get_bounds_cw(self):
        return _convert_se_to_cw(self._bounds)

    def _set_bounds_from_se(self, bounds_se):
        self._bounds = bounds_se

    def _set_bounds_from_cw(self, bounds_cw):
        self._bounds = _convert_cw_to_se_inplace(bounds_cw)

    def _add_offset(self, offset_each_dim):
        np.add(self._bounds, offset_each_dim[..., np.newaxis], out=self._bounds)


class NdBoundsCW(BoundFormatHandler):
    """ Handler for the format (center, width) """

    def __init__(self, bounds):
        super().__init__(bounds)

    def _get_bounds_se(self):
        return _convert_cw_to_se(self._bounds)

    def _get_bounds_cw(self):
        return self._bounds

    def _set_bounds_from_se(self, bounds_se):
        self._bounds = _convert_se_to_cw_inplace(bounds_se)

    def _set_bounds_from_cw(self, bounds_cw):
        self._bounds = bounds_cw

    def _add_offset(self, offset_each_dim):
        np.add(self._bounds[..., 0], offset_each_dim, out=self._bounds[..., 0])



# Copyright (C) 2021 Matthias Nadig

import numpy as np

from ._bounds import NdBounds
from ._bounds import NdBoundsSE, NdBoundsCW


def as_bounds_se(bounds,
                 require_n_dims=None, assert_order=True):
    """
    Assemble NdBounds object from an array or converts NdBounds to correct subtype
        --> receives array of format (start, end) or converts NdBounds accordingly
    """
    return _as_bounds(bounds, NdBoundsSE,
                      require_n_dims=require_n_dims,
                      assert_order=assert_order)


def as_bounds_cw(bounds,
                 require_n_dims=None):
    """
    Assemble NdBounds object from an array or converts NdBounds to correct subtype
        --> receives array of format (center, width) or converts NdBounds accordingly
    """
    return _as_bounds(bounds, NdBoundsCW,
                      require_n_dims=require_n_dims)


def _as_bounds(bounds, target_class,
               require_n_dims=None, assert_order=True):
    # Proceed depending on input type
    if isinstance(bounds, (np.ndarray, list, tuple)):
        # Input is an array containing the bounds, wrap it by the NdBounds object
        bounds = np.asarray(bounds)

        # Assert bounds array to be of shape (..., n_dims, 2)
        if bounds.ndim < 3 or bounds.shape[-1] != 2:
            raise ValueError('Bounds array has bad shape: (..., n_dims, 2) != {}'.format(bounds.shape))

        # Assert n_dims to be of certain value (dimensionality of bounds themselves, not the bounds array)
        if require_n_dims is not None:
            if bounds.shape[-2] != require_n_dims:
                raise ValueError(
                    'Given bounds do not fulfil user requirements. ' +
                    'Requested {}-dimensional bounds, got n_dims = {}'.format(require_n_dims, bounds.shape[-2]))

        # Check order of start and end (only in case bounds of type start-end are required)
        if assert_order and target_class == NdBoundsSE:
            is_wrong_order = np.less(np.subtract(bounds[..., 1], bounds[..., 0]), 0)
            if is_wrong_order.any():
                str_count = '{} of {} bounds'.format(np.sum(is_wrong_order), is_wrong_order.size)
                raise ValueError('Bounds have wrong order: start > end for {}'.format(str_count))

        bounds = NdBounds(target_class(bounds))
    elif isinstance(bounds, NdBounds):
        # Input is already of type NdBounds, convert to correct subtype if necessary
        if isinstance(bounds, target_class):
            # Type is already good, no conversion necessary
            pass
        elif target_class == NdBoundsSE:
            # Convert to required type: start-end
            bounds = NdBounds(target_class(bounds.get_bounds_se()))
        elif target_class == NdBoundsCW:
            # Convert to required type: center-width
            bounds = NdBounds(target_class(bounds.get_bounds_cw()))
        else:
            raise RuntimeError('Child of NdBounds has unexpected type: {}'.format(type(bounds)))
    else:
        raise ValueError('Unexpected type of bounds: {}'.format(bounds))

    return bounds
